/*				DB/C 9.1							*/
/*			DB/C Interface to ODBC					*/
/*		(c) Copyright  1997  Subject, Wills & Company	*/

#if 0
/*** tlib * release 9.1 version %v %f * */
/* release 9.1 version 3 17-Oct-97,09:07:04 */

/*** CHANGE HISTORY ***/
/* 1 SQLAODBC.C 10-Jun-96,17:15:30,`JSF' Release 9.1 copied from release 9.0 */
/* 2 SQLAODBC.C 17-Oct-97,06:53:44,`JSF' added version 9.0 changes           */
/* 3 SQLAODBC.C 17-Oct-97,09:07:04,`JSF' fixed odbc sdk includes             */
/*** CHANGE HISTORY ***/
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <windows.h>

#include <sql.h>  /* From ODBC SDK */
#include <sqlext.h>  /* From ODBC SDK */

#define INC_UTIL
#include "dbcsql.h"

#define MAXCURSORS 50

/* Cursor stucture */
struct cursearr {
        HSTMT hstmt;
        char cursorname[20];
        SWORD nResultCols;
        UCHAR **szColData;
        SDWORD *dwLen;
        UDWORD *udwColPrec;
};

static struct cursearr cursor_array[MAXCURSORS];


static HCURSOR hOldCursor;
static char errbuf[CMDSIZE];

/* Error variables */
static char szErrText[SQL_MAX_MESSAGE_LENGTH+1];
static char szErrState[SQL_SQLSTATE_SIZE+1];
static SWORD wErrMsgLen;
static SDWORD dwNativeErrCode;

/* Handles */
static HENV henv;
static HDBC hdbc;

static void dbcsqlerr(HENV, HDBC, HSTMT);


/* GETCURSORNUM */
int getcursornum(char *cursorname)
{
        int cursornum;
        
        for (cursornum = 0; cursornum < MAXCURSORS; cursornum++) {
        	if (!strcmp(cursor_array[cursornum].cursorname, cursorname)) {
			break;
		}
	}
        if (cursornum == MAXCURSORS) {
                /* New cursor so find an empty slot */
                for (cursornum = 0; cursornum < MAXCURSORS; cursornum++) {
                        if (!cursor_array[cursornum].cursorname[0]) break;
                }
                if (cursornum == MAXCURSORS) {
                        return(-1);
                }
                strcpy(cursor_array[cursornum].cursorname, cursorname);
        }
	return(cursornum);
}


/* DBCCONNECT */
int dbcconnect(char *cmdbuf, char *uid, char *passwd, char *server, char *string)
{
	int retcode;
	SWORD lenptr;

	if ((retcode = SQLAllocEnv(&henv)) != SQL_SUCCESS) {
		hdbc = SQL_NULL_HDBC;
		dbcsqlerr(henv, SQL_NULL_HDBC, SQL_NULL_HSTMT);
		return(retcode);
	}
	if ((retcode = SQLAllocConnect(henv, (HDBC FAR *)&hdbc)) != SQL_SUCCESS) {
		dbcsqlerr(henv, hdbc, SQL_NULL_HSTMT);
		return(retcode);
	}

	SQLSetConnectOption(hdbc, SQL_ODBC_CURSORS, SQL_CUR_USE_IF_NEEDED);

	hOldCursor = SetCursor(LoadCursor(NULL, IDC_WAIT));

	if (!uid[0] || !passwd[0] || !server[0]) {
		retcode = SQLDriverConnect(hdbc, GetActiveWindow(),
			(UCHAR FAR *)"", 0,
			(UCHAR FAR *)cmdbuf, CMDSIZE, &lenptr,
			SQL_DRIVER_COMPLETE_REQUIRED);
	} else {
		retcode = SQLConnect(hdbc,
			(UCHAR FAR *)server, SQL_NTS,
			(UCHAR FAR *)uid, SQL_NTS,
			(UCHAR FAR *)passwd, SQL_NTS);
	}

	if (retcode != SQL_SUCCESS) {
		dbcsqlerr(henv, hdbc, SQL_NULL_HSTMT);
		if (retcode != SQL_SUCCESS_WITH_INFO) {
			SQLFreeConnect(hdbc);
		}
	}
	
	SetCursor(hOldCursor);

	return(retcode);
}


/* DBCCLOSE */
int dbcclose(char *cmdbuf, int cursornum)
{
	HSTMT *stmtptr;
	SWORD ncols;
	struct cursearr *cu;
	int loop;

	cu = &cursor_array[cursornum];

	stmtptr = &cu->hstmt;

	if (*stmtptr != NULL) {
		SQLFreeStmt(*stmtptr, SQL_DROP);
		*stmtptr = NULL;
	}

	cu->cursorname[0] = 0;

	ncols = cu->nResultCols;

	/* free any previous allocations */
	if (cu->szColData != NULL) {
		for (loop = 0; loop < ncols; loop++) {
			free(cu->szColData[loop]);
		}
		free(cu->szColData);
		free(cu->udwColPrec);
		free(cu->dwLen);
		cu->szColData = NULL;
	}
	return(0);
}


/* DBCDISCONNECT */
int dbcdisconnect(void)
{
	SQLDisconnect(hdbc);
	SQLFreeConnect(hdbc);
	SQLFreeEnv(henv);
	hdbc = henv = NULL;
	return(0);
}


/* DBCCOMMIT */
int dbccommit(void)
{
	int retcode;

	retcode = SQLTransact(henv, hdbc, SQL_COMMIT);
	if (retcode != SQL_SUCCESS) {
		dbcsqlerr(henv, hdbc, SQL_NULL_HSTMT);
	}
	return(retcode);
}


/* DBCROLLBACK */
int dbcrollback(void)
{
	int retcode;

	retcode = SQLTransact(henv, hdbc, SQL_ROLLBACK);
	if (retcode != SQL_SUCCESS) {
		dbcsqlerr(henv, hdbc, SQL_NULL_HSTMT);
	}
	return(retcode);
}


/* DBCFETCH */
int dbcfetch(char *cmdbuf, int cursornum, int fetchtype, int index, unsigned char **destarr)
{
	SWORD ncols;
	HSTMT *stmtptr;
	UCHAR **dbcvar;
	int fnotfound;
	int loop;
	struct cursearr *cu;
	int retcode = 0;

	cu = &cursor_array[cursornum];

	ncols = cu->nResultCols;
	stmtptr = &cu->hstmt;

	if (ncols) {
		fnotfound = FALSE;
		switch (fetchtype) {
			case DBC_FET_PREV:
				retcode = SQLExtendedFetch(*stmtptr, SQL_FETCH_PREV,
					0, NULL, NULL);
				break;
			case DBC_FET_FIRST:
				retcode = SQLExtendedFetch(*stmtptr, SQL_FETCH_FIRST,
					0, NULL, NULL);
				break;
			case DBC_FET_LAST:
				retcode = SQLExtendedFetch(*stmtptr, SQL_FETCH_LAST,
					0, NULL, NULL);
				break;
			case DBC_FET_REL:
				retcode = SQLExtendedFetch(*stmtptr, SQL_FETCH_RELATIVE,
					index, NULL, NULL);
				break;
			case DBC_FET_ABS:
				retcode = SQLExtendedFetch(*stmtptr, SQL_FETCH_ABSOLUTE,
					index, NULL, NULL);
				break;
			case DBC_FET_NEXT:
				retcode = SQLExtendedFetch(*stmtptr, SQL_FETCH_NEXT,
					0, NULL, NULL);
				break;
		}
		dbcvar = destarr;
		if (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO) {
			if (retcode == SQL_SUCCESS_WITH_INFO) {
				dbcsqlerr(henv, hdbc, *stmtptr);
			}
			for (loop = 0; loop < ncols && dbcvar[loop] != NULL; loop++) {
				if (cu->dwLen[loop] == SQL_NULL_DATA) {
#ifdef ALLOW_NULL
					makenull(dbcvar[loop]);
#else
					cu->dwLen[loop] = 0;
					strtodbc((char *)cu->szColData[loop], 0, dbcvar[loop]);
#endif
				} else {
					strtodbc((char *)cu->szColData[loop], (long)cu->dwLen[loop], dbcvar[loop]);
				}
			}
		} else if (retcode < 0) {
			dbcsqlerr(henv, hdbc, *stmtptr);
		}
	} else {
		fnotfound = TRUE;
	}
	return(((fnotfound) ? 100 : (int)retcode));

}


/* DBCSELECT */
int dbcselect(char *cmdbuf, int cursornum, int scroll, int lock)
{
	HSTMT *stmtptr;
	SWORD ncols;
	SWORD nColNameLen, nColType, nScale, fNullable;
	char frealloc;
	int retcode;
	int retcode2;
	int maxcol;
	int loop;
	UDWORD odbcscroll, odbclock;
	struct cursearr *cu;

	cu = &cursor_array[cursornum];	

	hOldCursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
		
	stmtptr = &cu->hstmt;
	ncols = cu->nResultCols;

	/* Was this statement/cursor already allocated ? */
	frealloc = (*stmtptr != NULL);
	if (frealloc) {
		SQLFreeStmt(*stmtptr, SQL_DROP);
		retcode = SQLAllocStmt(hdbc, stmtptr);
		if (retcode == SQL_ERROR || retcode == SQL_INVALID_HANDLE) {
			dbcsqlerr(henv, hdbc, SQL_NULL_HSTMT);
			SetCursor(hOldCursor);
			return(-1);
		}
	} else {
		retcode = SQLAllocStmt(hdbc, stmtptr);
		if (retcode == SQL_ERROR || retcode == SQL_INVALID_HANDLE) {
			dbcsqlerr(henv, hdbc, SQL_NULL_HSTMT);
			SetCursor(hOldCursor);
			return(-1);
		}
	}
	/* free any previous allocations */
	if (cu->szColData != NULL) {
		for (loop = 0; loop < ncols; loop++) {
			free(cu->szColData[loop]);
		}
		free(cu->szColData);
		free(cu->udwColPrec);
		free(cu->dwLen);
		cu->szColData = NULL;
	}


	/* Set cursor name */
	retcode = SQLSetCursorName(*stmtptr, cu->cursorname, SQL_NTS);
	if (retcode == SQL_ERROR || retcode == SQL_INVALID_HANDLE) {
		dbcsqlerr(henv, hdbc, *stmtptr);
		SetCursor(hOldCursor);
		return(-1);
	}

	/* cursor options */
	if (lock) {
		switch (lock) {
			case DBC_CON_OPTTIME:
				odbclock = SQL_CONCUR_ROWVER;
				break;
			case DBC_CON_OPTVAL:
				odbclock = SQL_CONCUR_VALUES;
				break;
			case DBC_CON_RDONLY:
				odbclock = SQL_CONCUR_READ_ONLY;
				break;
			default:				/* DBC_CON_LOCK */
				odbclock = SQL_CONCUR_LOCK;
				break;
		}
		retcode = SQLSetStmtOption(*stmtptr, SQL_CONCURRENCY, odbclock);
		if (retcode == SQL_ERROR || retcode == SQL_INVALID_HANDLE) {
			dbcsqlerr(henv, hdbc, *stmtptr);
			SetCursor(hOldCursor);
			return(-1);
		}
	}
	if (scroll) {
		switch (scroll) {
			case DBC_CUR_FOR:
				odbcscroll = SQL_CURSOR_FORWARD_ONLY;
				break;
			case DBC_CUR_KEY:
				odbcscroll = SQL_CURSOR_KEYSET_DRIVEN;
				break;
			default:				/* DBC_CUR_DYN */
				odbcscroll = SQL_CURSOR_DYNAMIC;
				break;
		}
		retcode = SQLSetStmtOption(*stmtptr, SQL_CURSOR_TYPE, odbcscroll);
		if (retcode == SQL_ERROR || retcode == SQL_INVALID_HANDLE) {
			dbcsqlerr(henv, hdbc, *stmtptr);
			SetCursor(hOldCursor);
			return(-1);
		}
	}
	/* Execute the Select */
	retcode = SQLExecDirect(*stmtptr, (UCHAR FAR *)cmdbuf, SQL_NTS);
	if (retcode != SQL_SUCCESS) {
		dbcsqlerr(henv, hdbc, *stmtptr);
	}
	if (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO) { 
		/* Allocate memory for a single result row */			
		SQLNumResultCols(*stmtptr, &ncols);
		cu->nResultCols = ncols;
		if (ncols > 0) {
			cu->szColData = (UCHAR **)malloc(ncols * sizeof(UCHAR *));
			cu->udwColPrec = (UDWORD *)malloc(ncols * sizeof(UDWORD));
			cu->dwLen = (SDWORD *)malloc(ncols * sizeof(SDWORD));
		}
			
		/* Bind result columns to work variables */
		for (loop = 0; loop < ncols; loop++) {
			/* Get the scale info for column nCol */
			retcode2 = SQLDescribeCol(*stmtptr,
					(UWORD) (loop + 1),
					NULL,
					0,
					&nColNameLen,
					&nColType,
					&cu->udwColPrec[loop],
					&nScale,
					&fNullable);
			if (retcode2 == SQL_ERROR) {
				dbcsqlerr(henv, hdbc, *stmtptr);
			}
				
			/* Allocate work variable for column loop */
			if (nColType == SQL_CHAR || nColType == SQL_VARCHAR || nColType == SQL_LONGVARCHAR) {
				maxcol = cu->udwColPrec[loop] + 3;
			} else if (nColType == SQL_REAL || nColType == SQL_FLOAT || nColType == SQL_DOUBLE || nColType == SQL_TIMESTAMP) {
				maxcol = cu->udwColPrec[loop] + 10;
			} else if (nColType == SQL_BINARY || nColType == SQL_VARBINARY || nColType == SQL_LONGVARBINARY) {
				maxcol = (cu->udwColPrec[loop] * 2) + 4;
			} else {
				maxcol = cu->udwColPrec[loop] + 4;
			}

			cu->szColData[loop] = (UCHAR *)malloc(sizeof(UCHAR) * maxcol);
				
			/* Bind column nCol to work variable */
			retcode2 = SQLBindCol(*stmtptr,
					(UWORD) (loop + 1),
					SQL_C_CHAR,
					cu->szColData[loop],
					maxcol -1,
					&cu->dwLen[loop]);
			if (retcode2 == SQL_ERROR) {
				dbcsqlerr(henv, hdbc, *stmtptr);
			}
		}
	} else {
		cu->nResultCols = 0;
	}
	SetCursor(hOldCursor);
	return(retcode);
}


/* DBCUPDDEL */
int dbcupddel(char *cmdbuf, int fcursor, int cursornum, int fupdel)
{
	HSTMT stmt;
	int retcode;


	if (!fcursor) {
		hOldCursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
	}


	retcode = SQLAllocStmt(hdbc, &stmt);	
	if (retcode == SQL_ERROR || retcode == SQL_INVALID_HANDLE) {
		dbcsqlerr(henv, hdbc, SQL_NULL_HSTMT);
		return(-1);
	}

	retcode = SQLExecDirect(stmt, (UCHAR FAR *)cmdbuf, SQL_NTS);

	if (retcode != SQL_SUCCESS) {
		dbcsqlerr(henv, hdbc, stmt);
	}

	if (!fcursor) {
		SetCursor(hOldCursor);
	}
	SQLFreeStmt(stmt, SQL_DROP);
	return(retcode);
}


/* DBCEXECDIR */
int dbcexecdir(char *cmdbuf)
{

	HSTMT stmt;
	int retcode;

	hOldCursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
		
	/* Allocate a statement handle for sql command */
	retcode = SQLAllocStmt(hdbc, &stmt);
	if (retcode == SQL_ERROR || retcode == SQL_INVALID_HANDLE) {
		dbcsqlerr(henv, hdbc, SQL_NULL_HSTMT);
		SetCursor(hOldCursor);
		return(-1);
	}	
	/* Execute the statement */
	retcode = SQLExecDirect(stmt, (UCHAR FAR *)cmdbuf, SQL_NTS);
	if (retcode != SQL_SUCCESS) {
		dbcsqlerr(henv, hdbc, stmt);
	}

	SQLFreeStmt(stmt, SQL_DROP);

	SetCursor(hOldCursor);

	return(retcode);
}


/* DBCGETERR */
char *dbcgeterr(long code)
{
	char temp[20];
	
	if (!wErrMsgLen) {
		errbuf[0] = 0;
	} else {
		strcpy(errbuf, szErrState);
		strcat(errbuf, ": ");
		strcat(errbuf, (char *)szErrText);
		if (dwNativeErrCode) {
			strcat(errbuf, ": (#");
			sprintf(temp, "%ld", dwNativeErrCode);
			strcat(errbuf, temp);
			strcat(errbuf, ")");
			dwNativeErrCode = 0;
		}
		wErrMsgLen = 0;
	}
	return(errbuf);
}

static void dbcsqlerr(HENV henv, HDBC localcnct, HSTMT stmt)
{
	int retcode;
	
	retcode = SQLError(henv, localcnct, stmt,
		(UCHAR FAR *)szErrState,
		(SDWORD FAR *)&dwNativeErrCode,
		(UCHAR FAR *)szErrText, SQL_MAX_MESSAGE_LENGTH-1,
		&wErrMsgLen);
}

#ifdef WIN16
void clearglobal(void)
{
	memset(cursor_array, 0, MAXCURSORS * sizeof(struct cursearr));
	return;
}
#endif
