/*				DB/C 9.0							*/
/*			DB/C Interface to SYBASE					*/
/*		(c) Copyright  1994  Subject, Wills & Company	*/

#if 0
/*** tlib * release 9.0 version %v %f * */
/* release 9.0 version 5 23-Feb-96,18:06:40 */

/*** CHANGE HISTORY ***/
/* 1 SQLASYB.C 20-Mar-95,12:27:14,`SEN' Initial Release                      */
/* 2 SQLASYB.C 29-Mar-95,09:27:24,`SEN' Fixed some return codes              */
/* 3 SQLASYB.C 31-Mar-95,14:13:48,`SEN' Changed Error from dbinit            */
/* 4 SQLASYB.C 15-Feb-96,14:00:12,`SEN' Added data types for SQL Server      */
/* 5 SQLASYB.C 23-Feb-96,18:06:40,`SEN' Added casts to err hanler calls      */
/*** CHANGE HISTORY ***/
#endif


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#if defined(WIN32) | defined (WIN16)
#include <windows.h>
#endif

#define INC_UTIL
#include "dbcsql.h"

#include <sybfront.h>
#include <sybdb.h>
#include <syberror.h>

#define MAXCURSORS 10

/* Cursor stucture */
struct cursearr {
	DBCURSOR *hc;
	char cursorname[20];
	DBINT ncols;
	DBINT fetstatus;
	unsigned char **coldata;
	DBINT *collenres;			/* Amount Returned in Data Buffer */
	DBINT *collenout;			/* Size of Data Buffer */		
};

static struct cursearr cursor_array[MAXCURSORS];

/* Buffer for error information */
static char errbuf[CMDSIZE];

/* Library Initialization Flag */
static int initflag;

/* Error Variables */
static int errsz;

/* Handles */
LOGINREC *login;
DBPROCESS *dbproc;					

/* Prototypes */
int dbcsqlerr(DBPROCESS *, int, int, int, char *, char *);
int dbcsqlmsg(DBPROCESS *, DBINT, int, int, char *, char *, char *, int);

/* GETCURSORNUM */
int getcursornum(cursorname)
char *cursorname;
{
	int cursornum;
        
	for (cursornum = 0; cursornum < MAXCURSORS; cursornum++) {
		if (!strcmp(cursor_array[cursornum].cursorname, cursorname)) {
			break;
		}
	}
	if (cursornum == MAXCURSORS) {
		/* New cursor so find an empty slot */
		for (cursornum = 0; cursornum < MAXCURSORS; cursornum++) {
			if (!cursor_array[cursornum].cursorname[0]) {
				break;
			}
		}
		if (cursornum == MAXCURSORS) {
			return(-1);
		}
		strcpy(cursor_array[cursornum].cursorname, cursorname);
	}
	return(cursornum);
}

/* DBCCONNECT */
int dbcconnect(cmdbuf, uid, passwd, server, string)
char *cmdbuf;
char *uid;
char *passwd;
char *server;
char *string;
{
	RETCODE sybretcode;
	int retcode;

	errsz = 0;

	if (!initflag) {
		sybretcode = dbinit();
		if (sybretcode == FAIL) {
			strcpy(errbuf, "[DB/C]: Unable to initialize DB-Library");
			errsz = 31;
			return(-1);
		}
		initflag = 1;
	}

	dbsetversion(DBVERSION_100);

	dberrhandle((EHANDLEFUNC)dbcsqlerr);
	dbmsghandle((MHANDLEFUNC)dbcsqlmsg);

	login = dblogin();
	if (login == NULL) {
		return(-1);
	}
	if (uid[0]) {
		DBSETLUSER(login, uid);
	}
	if (passwd[0]) {
		DBSETLPWD(login, passwd);
	}
	if (string[0]) {
		DBSETLHOST(login, string);
	}
	if (dbproc) {
		dbclose(dbproc);
		dbproc = NULL;
	}

	if (server[0]) {
		dbproc = dbopen(login, server);
	} else {
		dbproc = dbopen(login, NULL);
	}
	dbloginfree(login);
	if (dbproc == NULL) {
		return(-1);
	}
	return(0);
}

/* DBCCLOSE */
int dbcclose(cmdbuf, cursornum)
char *cmdbuf;
int cursornum;
{
	struct cursearr *cu;

	cu = &cursor_array[cursornum];
	errsz = 0;

	dbcursorclose(cu->hc);
	cu->hc = NULL;
	cu->cursorname[0] = 0;
	return(0);
}

/* DBCDISCONNECT */
int dbcdisconnect()
{
	errsz = 0;
	if (dbproc) {
		dbclose(dbproc);
		dbproc = NULL;
	}
	return(0);
}

/* DBCCOMMIT */
int dbccommit()
{
	RETCODE sybretcode;

	errsz = 0;

	sybretcode = dbcmd(dbproc, "COMMIT TRANSACTION");
	if (sybretcode == FAIL) {
		return(-1);
	}
	sybretcode = dbsqlexec(dbproc);
	if (sybretcode == FAIL) {
		return(-1);
	}
	sybretcode = dbresults(dbproc);
	if (sybretcode == FAIL) {
		return(-1);
	}
	return(0);
}

/* DBCROLLBACK */
int dbcrollback()
{
	RETCODE sybretcode;

	errsz = 0;

	sybretcode = dbcmd(dbproc, "ROLLBACK TRANSACTION");
	if (sybretcode == FAIL) {
		return(-1);
	}
	sybretcode = dbsqlexec(dbproc);
	if (sybretcode == FAIL) {
		return(-1);
	}
	sybretcode = dbresults(dbproc);
	if (sybretcode == FAIL) {
		return(-1);
	}
	return(0);
}

/* DBCFETCH */
int dbcfetch(cmdbuf, cursornum, fetchtype, index, destarr)
char *cmdbuf;
int cursornum;
int fetchtype;
int index;
unsigned char **destarr;
{
	RETCODE sybretcode;
	DBINT ncols;
	unsigned char **dbcvar;
	int fnotfound;
	int loop;
	int reslen;
	struct cursearr *cu;

	cu = &cursor_array[cursornum];

	errsz = 0;

	ncols = cu->ncols;

	if (ncols) {
		fnotfound = 0;
		switch (fetchtype) {
			case DBC_FET_PREV:
				sybretcode = dbcursorfetch(cu->hc, (DBINT)FETCH_PREV, (DBINT)0);
				break;
			case DBC_FET_FIRST:
				sybretcode = dbcursorfetch(cu->hc, (DBINT)FETCH_FIRST, (DBINT)0);
				break;
			case DBC_FET_LAST:
				sybretcode = dbcursorfetch(cu->hc, (DBINT)FETCH_LAST, (DBINT)0);
				break;
			case DBC_FET_REL:
				sybretcode = dbcursorfetch(cu->hc, (DBINT)FETCH_RELATIVE, (DBINT)index);
				break;
			case DBC_FET_ABS:
				sybretcode = dbcursorfetch(cu->hc, (DBINT)FETCH_RANDOM, (DBINT)index);
				break;
			case DBC_FET_NEXT:
				if (cu->fetstatus & FTC_ENDOFRESULTS) {
					cu->fetstatus = FTC_MISSING;
				} else {
					sybretcode = dbcursorfetch(cu->hc, (DBINT)FETCH_NEXT, (DBINT)0);
				}
				break;
		}
		dbcvar = destarr;
		if (sybretcode == SUCCEED && cu->fetstatus & FTC_SUCCEED) {
			for (loop = 0; loop < ncols && dbcvar[loop] != NULL; loop++) {
				reslen = (int)strlen((char *)cu->coldata[loop]);
				if (reslen == 0) {
#ifdef ALLOW_NULL
					makenull(dbcvar(loop));
#else
					strtodbc((char *)cu->coldata[loop], 0, dbcvar[loop]);
#endif
				} else {
					strtodbc((char *)cu->coldata[loop], (long)reslen, dbcvar[loop]);
				}
			}
		} else {
			return(100);
		}
	} else {
		return(100);
	}
	if (sybretcode == FAIL) {
		return(-1);
	}
	return(0);		
}

/* DBCSELECT */
int dbcselect(cmdbuf, cursornum, scroll, lock)
char *cmdbuf;
int cursornum;
int scroll;
int lock;
{
	DBINT ncols;
	int frealloc;
	int maxcol;
	DBINT loop;
	DBINT natcoltype;
	DBINT natcollen;
	DBINT cursorcon;
	DBINT cursortype;
	RETCODE sybretcode;
	struct cursearr *cu;

	cu = &cursor_array[cursornum];

	errsz = 0;

	ncols = cu->ncols;
	cu->ncols = 0;
	if (cu->hc == NULL) {
		frealloc = 0;
	} else {
		frealloc = 1;
	}
	if (frealloc) {
		dbcursorclose(cu->hc);
		cu->hc = NULL;
	}

	/* free previous allocs */
	if (cu->coldata != NULL) {
		for (loop = 0; loop < ncols; loop++) {
			if (cu->coldata[loop]) {
				free(cu->coldata[loop]);
			} else {
				break;
			}
		}
		free(cu->coldata);
		free(cu->collenres);
		free(cu->collenout);
		cu->coldata = NULL;
	}

	/* set flags for scrolling and concurrency */
	switch (lock) {
		case DBC_CON_OPTTIME:
			cursorcon = CUR_OPTCC;
			break;
		case DBC_CON_OPTVAL:
			cursorcon = CUR_OPTCCVAL;
			break;
		case DBC_CON_RDONLY:
			cursorcon = CUR_READONLY;
			break;
		default:				/* DBC_CON_LOCK */
			cursorcon = CUR_LOCKCC;
			break;
	}
	switch (scroll) {
		case DBC_CUR_FOR:
			cursortype = CUR_FORWARD;
			break;
		case DBC_CUR_KEY:
			cursortype = CUR_KEYSET;
			break;
		default:				/* DBC_CUR_DYN */
			cursortype = CUR_DYNAMIC;
			break;
	}
	
	/* open cursor */
	cu->hc = dbcursoropen(dbproc, (BYTE *)cmdbuf, cursortype,
						cursorcon, (DBINT) 1,
						(DBINT *)&cu->fetstatus);
	if (cu->hc == NULL) {
		return(-1);
	}

	cu->fetstatus = 0;

	/* bind variables */
	sybretcode = dbcursorinfo(cu->hc, &ncols, (DBINT *)&frealloc);
	if (sybretcode == FAIL) {
		return(-1);
	}
	cu->ncols = ncols;
	if (ncols > 0) {
		cu->coldata = (unsigned char **)malloc(ncols * sizeof(unsigned char *));
		cu->collenres = (DBINT *)malloc(ncols * sizeof(DBINT));
		cu->collenout = (DBINT *)malloc(ncols * sizeof(DBINT));
	}
	for (loop = 0; loop < ncols; loop++) {
		sybretcode = dbcursorcolinfo(cu->hc, loop + 1, NULL,
							&natcoltype,  &natcollen, NULL);
		if (sybretcode == FAIL) {
			return(-1);
		}
		switch (natcoltype) {
			case SYBINT1:
				maxcol = 4;
				break;
			case SYBINT2:
				maxcol = 7;
				break;
			case SYBINT4:
			case SYBINTN:
				maxcol = 12;
				break;
			case SYBNUMERIC:
			case SYBDECIMAL:
				maxcol = (natcollen * 2) + 10;
				break;
			case SYBDATETIME:
			case SYBDATETIME4:
			case SYBDATETIMN:
			case SYBMONEY:
			case SYBFLT8:
			case SYBFLTN:
				maxcol = 25;
				break;
			case SYBMONEY4:
			case SYBMONEYN:
			case SYBREAL:
				maxcol = 16;
				break;
			case SYBCHAR:
			case SYBVARCHAR:
				maxcol = natcollen + 1;
				break;
			case SYBBINARY:
				maxcol = natcollen + 1;
				break;
			case SYBBIT:
				maxcol = 2;
				break;
			default:
				maxcol = 0;
				return(-1);
		}
		cu->coldata[loop] = (unsigned char *)malloc(sizeof(unsigned char) * maxcol);
		
		sybretcode = dbcursorbind(cu->hc, (int)(loop + 1),
							NTBSTRINGBIND, maxcol,
							&cu->collenres[loop],
							cu->coldata[loop],
							NULL);
		if (sybretcode == FAIL) {
			return(-1);
		}
	}
	return(0);
}

/* DBCUPDDEL */
int dbcupddel(cmdbuf, fcursor, cursornum, fupddel)
char *cmdbuf;
int fcursor;
int cursornum;
int fupddel;
{
	RETCODE sybretcode;
	char *ptr1;
	int loop;
	struct cursearr *cu;

	cu = &cursor_array[cursornum];

	errsz = 0;

	if (fcursor) {
		if (fupddel == DBC_DEL) {
			sybretcode = dbcursor(cu->hc, CRS_DELETE, 1, NULL, NULL);
			if (sybretcode == FAIL) {
				return(-1);
			}
		} else {
			ptr1 = cmdbuf + strlen(cmdbuf) - 1;
			for (loop = 0; loop < 4; loop++) {
				while (*ptr1 != ' ' && *ptr1 != '\t') {
					ptr1--;
				}
				while (*ptr1 == ' ' || *ptr1 == '\t') {
					ptr1--;
				}
			}
			*(ptr1 + 1) = 0;
			sybretcode = dbcursor(cu->hc, CRS_UPDATE, 1, NULL, (BYTE *)cmdbuf);
			if (sybretcode == FAIL) {
				return(-1);
			}
		}
	} else {
		sybretcode = dbcmd(dbproc, cmdbuf);
		if (sybretcode == FAIL) {
			return(-1);
		}
		sybretcode = dbsqlexec(dbproc);
		if (sybretcode == FAIL) {
			return(-1);
		}
		sybretcode = dbresults(dbproc);
		if (sybretcode == FAIL) {
			return(-1);
		}
	}
	return(0);
}

/* DBCEXECDIR */
int dbcexecdir(cmdbuf)
char *cmdbuf;
{
	RETCODE sybretcode;

	errsz = 0;

	sybretcode = dbcmd(dbproc, cmdbuf);
	if (sybretcode == FAIL) {
		return(-1);
	}
	sybretcode = dbsqlexec(dbproc);
	if (sybretcode == FAIL) {
		return(-1);
	}
	while (TRUE) {
		sybretcode = dbresults(dbproc);
		if (sybretcode == NO_MORE_RESULTS) {
			break;
		}
		if (sybretcode == FAIL) {
			dbcancel(dbproc);
			return(-1);
			break;
		}
	}
	if (dbhasretstat(dbproc)) {
		return((int) dbretstatus(dbproc));
	} else {
		return(0);
	}
}

/* DBCGETERR */
char *dbcgeterr(code)
long code;
{
	if (!errsz) {
		errbuf[0] = 0;
	}
	errsz = 0;
	return(errbuf);
}

/* DBCSQLERR */
int dbcsqlerr(dbprocloc, severity, dberr, oserr, dberrstr, oserrstr)
DBPROCESS *dbprocloc;
int severity;
int dberr;
int oserr;
char *dberrstr;
char *oserrstr;
{
	char tempstr[20];
	int fprevmsg;

	if (!errsz) {
		errbuf[0] = 0;
	}
	if (dberr == SYBESMSG || dberr == SYBEPWD) {
		fprevmsg = 1;
	} else {
		fprevmsg = 0;
	}
	strcat(errbuf, "ERROR MESSAGE");
	if (!fprevmsg && severity) {
		sprintf(tempstr, "%d", severity);
		strcat(errbuf, "[Severity: ");
		strcat(errbuf, tempstr);
		strcat(errbuf, "]");
	}
	if (!fprevmsg && dberr) {
		sprintf(tempstr, "%d", dberr);
		strcat(errbuf, "[Err No: ");
		strcat(errbuf, tempstr);
		strcat(errbuf, "]");
	}
	if (!fprevmsg && dberrstr) {
		strcat(errbuf, "[Msg Text: ");
		strcat(errbuf, dberrstr);
		strcat(errbuf, "]");
	}
	if (oserr) {
		sprintf(tempstr, "%d", dberr);
		strcat(errbuf, "[OS Err No: ");
		strcat(errbuf, tempstr);
		strcat(errbuf, "]");
	}
	if (oserrstr) {
		strcat(errbuf, "[OS Msg Text: ");
		strcat(errbuf, oserrstr);
		strcat(errbuf, "]");
	}
	errsz = (int)strlen(errbuf);
	return(INT_CANCEL);
}

/* DBCSQLMSG */
int dbcsqlmsg(dbprocloc, msgno, msgstate, severity, msgtext, srvname, procname, line)
DBPROCESS *dbprocloc;
DBINT msgno;
int msgstate;
int severity;
char *msgtext;
char *srvname;
char *procname;
int line;
{
	char tempstr[20];
	int fprevmsg;

	if (!errsz) {
		errbuf[0] = 0;
	}
	strcat(errbuf, "SERVER MESSAGE");
	if (msgno) {
		sprintf(tempstr, "%d", msgno);
		strcat(errbuf, "[Msg No: ");
		strcat(errbuf, tempstr);
		strcat(errbuf, "]");
	}
	if (msgstate) {
		sprintf(tempstr, "%d", msgstate);
		strcat(errbuf, "[Msg State: ");
		strcat(errbuf, tempstr);
		strcat(errbuf, "]");
	}
	if (severity) {
		sprintf(tempstr, "%d", severity);
		strcat(errbuf, "[Severity: ");
		strcat(errbuf, tempstr);
		strcat(errbuf, "]");
	}
	if (srvname) {
		strcat(errbuf, "[Server Name: ");
		strcat(errbuf, srvname);
		strcat(errbuf, "]");
	}
	if (procname) {
		strcat(errbuf, "[Proc Name: ");
		strcat(errbuf, procname);
		if (line) {
			sprintf(tempstr, "%d", line);
			strcat(errbuf, "(Line No: ");
			strcat(errbuf, tempstr);
			strcat(errbuf, ")");
		}
		strcat(errbuf, "]");
	}
	if (msgtext) {
		strcat(errbuf, "[Msg Text: ");
		strcat(errbuf, msgtext);
		strcat(errbuf, "]");
	}
	errsz = strlen(errbuf);
	return(0);
}

