/*				DB/C 9.1			*/
/*			DB/C SQL Utility Functions		*/
/*		(c) Copyright  1997  Subject, Wills & Company	*/

/*  This module is provided for use with DB/C SQL modules.  The		*/
/*  routines can be used to convert DB/C variables to text strings	*/
/*  text strings to DB/C variables, and routines to expand an SQL	*/
/*  command with DB/C variables */

#if 0
/*** tlib * release 9.1 version %v %f * */
/* release 9.1 version 3 17-Oct-97,09:06:36 */

/*** CHANGE HISTORY ***/
/* 1 SQLUTIL.C 10-Jun-96,17:15:40,`JSF' Release 9.1 copied from release 9.0  */
/* 2 SQLUTIL.C 17-Oct-97,06:53:22,`JSF' added version 9.0 changes            */
/* 3 SQLUTIL.C 17-Oct-97,09:06:36,`JSF' fixed compiler warnings              */
/*** CHANGE HISTORY ***/
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#if defined(WIN32) | defined(WIN16)
#define memicmp(a, b, c) _memicmp(a, b, c)
#define fcvt(a, b, c, d) _fcvt(a, b, c, d)
#endif

#include "dbcsql.h"


#define FLAG_USING 1
#define FLAG_OF 2

#define VALUE_SINGLE 1
#define VALUE_DOUBLE 2

static char work[CMDSIZE];  /* work variable for build */
static char swork[12];
static int right;
static int left;
static int dec;
static int sign;
static double fvar;

int dbctostr(unsigned char *, char *, int);

/* Convert a DB/C variable to ascii string */
int dbctostr(dbcvar, str, maxlen)
unsigned char *dbcvar;
char *str;
int maxlen;
{
	int i1, i2;
	long lwork;
	char /*far*/ *ptr;
	

	if (dbcvar == NULL) {
		str[0] = 0;
		return(0);
	}
	if (dbcvar[0] < 0x80) {  /* small DIM */
		i1 = (int) dbcvar[0];
		if (i1) {
			i2 = (int)dbcvar[1] - i1 + 1;
			if (i2 > maxlen) return(-1);
			memcpy(str, &dbcvar[i1 + 2], i2);
		}
		else i2 = 0;
		str[i2] = 0;
	}
	else if (dbcvar[0] == 0xf0) {  /* large DIM */
		i1 = dbcvar[1] + (dbcvar[2] << 8);
		if (i1) {
			i2 = dbcvar[3] + (dbcvar[4] << 8);
			if (i2 > maxlen) return(-1);
			memcpy(str, &dbcvar[i1 + 6], i2);
		}
		else i2 = 0;
		str[i2] = 0;
	}
	else if ((dbcvar[0] & 0xe0) == 0x80) {  /* FORM */
		i2 = dbcvar[0] & 0x1f;
		if (i2 > maxlen) return(-1);
		memcpy(str, &dbcvar[1], i2);
		str[i2] = 0;
	}
	else if (dbcvar[0] == 0xfc) {  /* INT */
		memcpy((char *) &lwork, &dbcvar[2], 4);
		if (lwork < 0) {
			sign = 1;
			lwork = -lwork;
		}
		else sign = 0;
		swork[11] = 0;
		i1 = 11;
		do swork[--i1] = (char)(lwork % 10 + '0');
		while (lwork /= 10);
		if (sign) swork[--i1] = '-';
		strcpy(str, &swork[i1]);
	}
	else if ((dbcvar[0] & 0xfc) == 0xf8) {  /* FLOAT */
		left = (((dbcvar[0] & 0x3) << 3) + (dbcvar[1] >> 5));
		right = dbcvar[1] & 0x1f;
		if (right) i2 = left + right + 1;
		else i2 = left;
		if (i2 > maxlen) return(-1);
		memcpy ((char *) &fvar, &dbcvar[2], 8);
		ptr = (char *) fcvt(fvar, right, &dec, &sign);
		if (dec >= left) memcpy (str, &ptr[dec - left], left);
		else {
			memset(str, ' ', left - dec);
			memcpy(&str[left - dec], ptr, dec);
		}
		if (right) {
			str[left] = '.';
			memcpy(&str[left + 1], &ptr[dec], right);
		}
		if (sign) {
			for (i1 = 0; str[i1] == ' '; i1++);
			if (i1 != 0) str[i1 - 1] = '-';
		}
		str[i2] = 0;
	}
	else if (dbcvar[0] == 0xe0 || dbcvar[0] == 0xe1) {  /* literal */
		i2 = dbcvar[1];
		if (i2 > maxlen) return(-1);
		memcpy(str, &dbcvar[2], i2);
		str[i2] = 0;
	}
	else str[0] = 0;
	return(0);
}

/* Convert ASCII string to a DB/C variable */
void strtodbc(str, len, dbcvar)
char *str;
unsigned long len;
unsigned char *dbcvar;
{
	int i1, i2, lft1, lft2, rgt1, rgt2, numflg;
	long lwork;
	double x1;

	i1 = (int)len;
	if (dbcvar[0] < 0x80) {  /* small DIM */
		if (i1) {
			if ((unsigned long)dbcvar[2] < len) i1 = (int)dbcvar[2];
			memcpy(&dbcvar[3], str, i1);
			dbcvar[0] = 1;
			dbcvar[1] = (unsigned char) i1;
		}
		else
			dbcvar[0] = dbcvar[1] = 0;
	}
	else if (dbcvar[0] == 0xf0) {  /* large DIM */
		if (i1) {
			i2 = dbcvar[5] + (dbcvar[6] << 8);
			if (i2 < i1) i1 = i2;
			memcpy(&dbcvar[7], str, i1);
			dbcvar[1] = 1;
			dbcvar[2] = 0;
			dbcvar[3] = (unsigned char) i1;
			dbcvar[4] = (unsigned char)(i1 >> 8);
		}
		else dbcvar[1] = dbcvar[2] = dbcvar[3] = dbcvar[4] = 0;
	}
	else if ((dbcvar[0] & 0xe0) == 0x80) {  /* FORM */
		/* test for valid numeric field */
		for (lft1 = 0; str[lft1] == ' '; lft1++);
		if (str[lft1] == '-') lft1++;
		while (isdigit(str[lft1])) lft1++;
		if (str[lft1] == '.') {
			for (rgt1 = lft1 + 1; isdigit(str[rgt1]); rgt1++);
			rgt1 -= lft1;
		}
		else rgt1 = 0;
		if (i1 && lft1 + rgt1 == i1 && isdigit(str[i1 - 1])) numflg = 1;
		else numflg = 0;

		/* get size of form variable */
		for (i2 = dbcvar[0] & 0x1f, lft2 = 0; lft2 < i2 && dbcvar[lft2 + 1] != '.'; lft2++);
		if (lft2 < i2) rgt2 = i2 - lft2;
		else rgt2 = 0;

		/* zero form variable */
		memset(&dbcvar[1], ' ', lft2);
		if (rgt2) {
			dbcvar[lft2 + 1] = '.';
			memset(&dbcvar[lft2 + 2], '0', rgt2 - 1);
		}
		else dbcvar[lft2] = '0';

		/* move string to form variable */
		if (numflg) {
			i1 = i2 = 0;
			if (lft1 > lft2) i1 = lft1 - lft2;
			if (lft1 < lft2) i2 = lft2 - lft1;
			memcpy(&dbcvar[i2 + 1], &str[i1], lft2 - i2);
			if (rgt1 > rgt2) rgt1 = rgt2;
			if (rgt1) memcpy(&dbcvar[lft2 + 1], &str[lft1], rgt1);
		}
	}
	else if (dbcvar[0] == 0xfc) {  /* INT */
		lwork = atol(str);
		memcpy(&dbcvar[2], (char *) &lwork, 4);
	}
	else if ((dbcvar[0] & 0xfc) == 0xf8) {  /* FLOAT */
		x1 = atof(str);
		memcpy(&dbcvar[2], (char *) &x1, 8);
	}
}


/* Place a null value in a DB/C variable */
void makenull(dbcvar)
unsigned char *dbcvar;
{
	int i1, i2;
	int decflag = 0;

	if (dbcvar[0] < 0x80) {  /* small char */
		memset(&dbcvar[3], ' ', dbcvar[2]);
		dbcvar[0] = 0;
		dbcvar[1] = 0xFF;
	}
	else if (dbcvar[0] == 0xF0) {  /* large char */
		i2 = dbcvar[5] + (dbcvar[6] << 8);
		memset(&dbcvar[7], ' ', i2);
		dbcvar[1] = dbcvar[2] = 0;
		dbcvar[3] = dbcvar[4] = 0xFF;
	}
	else if ((dbcvar[0] >= 0x80 && dbcvar[0] < 0xA0) || (dbcvar[0] >= 0xF8 && dbcvar[0] <= 0xFC)) {  /* num, float, int */
		if (dbcvar[0] == 0xFC) { /* int */
			memset(&dbcvar[2], 0, 4);
		}
		else if (dbcvar[0] >= 0xF8) { /* float */
			memset(&dbcvar[2], 0, 8);
		}
		else {  /* form */
			i2 = dbcvar[0] & 0x1F;
			decflag = 0;
			for (i1 = 1; i1 <= i2; i1++) {
				if (dbcvar[i1] == '.') {
					decflag = i1;
					continue;
				}
				dbcvar[i1] = '0';
			}
			if (decflag) {
				for (i1 = 1; i1 < (decflag - 1); i1++) {
					dbcvar[i1] = ' ';
				}
			} else {
				for (i1 = 1; i1 < i2; i1++) {
					dbcvar[i1] = ' ';
				}
			}

		}
		if (dbcvar[0] < 0xA0) {
			if (dbcvar[1] != '.') dbcvar[1] = dbcvar[0] & 0x1F;
			else dbcvar[1] = dbcvar[0];
			dbcvar[0] = 0x80;
		}
		else if (dbcvar[0] == 0xFC) dbcvar[1] |= 0x80;
		else dbcvar[0] -= 0x04;
	}
	return;
}

 /* Remove USING <cursorname> from cmdstr */
 /* copy <cursorname> to cname */
 /* Return 0 if USING <cursorname> found, else Return 1 */
int getcursorname(cmdstr, cname)
char *cmdstr, *cname;
{
	int i1, len, upos;
	int flag;
	int valueflag;
	
	flag = 0;
	valueflag = 0;
	len = strlen(cmdstr);

	for (upos = 0; upos < len; upos++) {
		switch (valueflag) {
			case VALUE_SINGLE:
				if (cmdstr[upos] == '\'') {
					valueflag = 0;
				}
				break;
			case VALUE_DOUBLE:
				if (cmdstr[upos] == '"') {
					valueflag = 0;
				}
				break;
			default:
				switch (cmdstr[upos]) {
					case '\'':
						valueflag = VALUE_SINGLE;
						break;
					case '"':
						valueflag = VALUE_DOUBLE;
						break;
					default:
						if (toupper(cmdstr[upos]) == 'U' && !memicmp(&cmdstr[upos], "USING ", 6)) {
							flag = FLAG_USING;
						} else if (toupper(cmdstr[upos]) == 'O' && !memicmp(&cmdstr[upos], "OF ", 3)) {
							flag = FLAG_OF;
						}
						break;
				}
				break;
		}
		if (flag) {
			break;
		}
	}
	if (upos == len) {
		/* USING <cursorname> not found so use table_cursor */
		strcpy(cname, "table_cursor");
		i1 = 0;
	}
	else if (flag == FLAG_USING) {
		i1 = upos + 6;
		while(cmdstr[i1] == ' ') i1++;
		while (cmdstr[i1] && cmdstr[i1] != ' ') {
			*cname = cmdstr[i1];
			cname++;
			i1++;
		}
		*cname = 0;
		len = strlen(&cmdstr[i1]);
		if (len > 0) {
			memmove(&cmdstr[upos], &cmdstr[i1], len);
			cmdstr[upos + len] = 0;
		}
		else cmdstr[upos] = 0;
		i1 = 1;
	}
	else {
		i1 = upos + 3;
		while (cmdstr[i1] == ' ')
			i1++;
		while (cmdstr[i1] && cmdstr[i1] != ' ') {
			*cname = cmdstr[i1];
			cname++;
			i1++;
		}
		*cname = 0;
		len = strlen(&cmdstr[i1]);
		if (len > 0) {
			memmove(&cmdstr[upos], &cmdstr[i1], len);
			cmdstr[upos + len] = 0;
		}
		else cmdstr[upos] = 0;
		i1 = 1;
	}
	return(i1);
}


/* Variable substitution */
int buildsql(cmdline, arglist)
char *cmdline;
unsigned char *arglist[];
{
	int cmdpos, wrkpos, argcnt, num;
	int oe;

	for (argcnt = 0; arglist[argcnt] != NULL; argcnt++);
	cmdpos = wrkpos = 0;
	while (cmdline[cmdpos]) {
		if (cmdline[cmdpos] == '\\') {
			oe = 0;
			while (cmdline[cmdpos] == '\\') {
				cmdpos++;
				if (oe++ % 2) {
					work[wrkpos++] = '\\';
				}
			}
			if (cmdline[cmdpos] && (oe % 2)) {
				work[wrkpos++] = cmdline[cmdpos++];
			}
		} else if (cmdline[cmdpos] == ':' && isdigit(cmdline[cmdpos + 1])) {
			cmdpos++;
			for (num = 0; isdigit(cmdline[cmdpos]); ) {
				num = num * 10 + cmdline[cmdpos++] - '0';
			}
			num--;
			if (num >= 0 && num < argcnt) {
				if (dbctostr(arglist[num], &work[wrkpos], CMDSIZE) == -1) {
					return(-1);
				}
				wrkpos += strlen(&work[wrkpos]);
			}
		}
		else work[wrkpos++] = cmdline[cmdpos++];
		if (cmdpos >= CMDSIZE || wrkpos >= CMDSIZE) return(-1);
	}
	work[wrkpos] = 0;
	strcpy(cmdline, work);
	return(wrkpos);
}

